/*!
A submodule that contains the different tree node types as a single huge enumeration.

Copyright © 2020 Santtu Söderholm
*/
use super::*;
use crate::common::AplusActiveElementClear;
use crate::common::AplusActiveElementInputType;
use crate::common::AplusActiveElementOutputType;
use crate::common::AplusExerciseStatus;
use crate::common::AplusRadarTokenizer;
use crate::common::QuizPoints;
use crate::doctree::directives::AdmonitionType;


/// An enumeration of the different possible document node types.
///
/// Some of the nodes listed here are redundant. This is because
/// reStructuredText documentation also [lists](https://docutils.sourceforge.io/docs/ref/doctree.html#element-reference)
/// nodes that (multiply) inherit from other nodes in the [implementation](https://sourceforge.net/p/docutils/code/HEAD/tree/trunk/docutils/docutils/nodes.py),
/// but Rust has no concept of inheritance built in.
#[derive(Debug)]
pub enum TreeNodeType {

    /// The abbreviation element is an inline element used to represent an abbreviation being used in the document. An example of an abbreviation is 'St' being used instead of 'Street'.
    Abbreviation {
        name: Option<String>,
        classes: Option<String>,
    },

    /// A reference to a web address.
    AbsoluteURI { text: String },

    /// TODO
    Acronym,

    /// The address element holds the surface mailing address information for the author (individual or group) of the document, or a third-party contact address. Its structure is identical to that of the literal_block element: whitespace is significant, especially newlines.
    Address,

    /// Covers all of the standard admonition types of reStructuredText:
    ///
    /// 1. attention,
    /// 2. caution,
    /// 3. danger,
    /// 4. error,
    /// 5. hint,
    /// 6. important,
    /// 7. note,
    /// 8. tip and
    /// 9. warning,
    /// 10. admonition
    ///
    /// All of these have the same fields, and vary only in how they will be displayed.
    /// The last variant "admonition" may also contain a title. as one of its fields.
    Admonition {
        content_indent: usize,
        classes: Option<String>,
        name: Option<String>,
        variant: AdmonitionType,
    },

    /// An optional attribution of a `BlockQuote`.
    /// An attribution is a text block beginning with "--", "---", or a true em-dash,
    /// flush left within the block quote.
    /// If the attribution consists of multiple lines,
    /// the left edges of the second and subsequent lines must align.
    ///
    /// If a `BlockQuote` contains an attribution,
    /// the following node may be a `BlockQuote as well,
    /// but not otherwise (as it will be interpreted as a part of the previous quote).
    Attribution {
        raw_text: String,
        // parsed_text: Vec<TreeNodeType>
    },

    /// TODO
    Author,

    /// TODO
    Authors,

    /// The "sectnum" (or "section-numbering") directive automatically numbers sections and subsections in a document (if not disabled by the
    /// --no-section-numbering command line option or the sectnum_xform configuration setting).
    ///
    /// Section numbers are of the "multiple enumeration" form, where each level has a number, separated by periods. For example,
    /// the title of section 1, subsection 2, subsubsection 3 would have "1.2.3" prefixed.
    ///
    /// The "sectnum" directive does its work in two passes: the initial parse and a transform. During the initial parse, a "pending" element is
    /// generated which acts as a placeholder, storing any options internally.
    /// At a later stage in the processing, the "pending" element triggers a transform, which adds section numbers to titles. Section numbers are
    /// enclosed in a "generated" element, and titles have their "auto" attribute set to "1".
    ///
    /// Details: https://docutils.sourceforge.io/docs/ref/rst/directives.html#automatic-section-numbering
    AutomaticSectionNumbering {
        depth: Option<u32>,
        prefix: Option<String>,
        suffix: Option<String>,
        start: Option<u32>,
    },

    /// Text indented relative to previous text,
    /// without markup indicating the start of
    /// a list or other container nodes.
    /// A block quote may end with an `Attribution`,
    /// which allows placing multiple block quotes
    /// in a sequence.
    ///
    /// Also generated by the `epigraph`, `highlights` and `pull-quote` directives.
    BlockQuote { body_indent: usize },

    /// An unnumbered list node. These may only contain `BulletListItem` nodes
    /// or `EmptyLine`s as their direct children.
    BulletList {
        bullet: char,
        bullet_indent: usize,
        text_indent: usize,
    },

    /// An unnumbered list item. Can contain any `Body` level elements as its direct children.
    BulletListItem {
        bullet: char,
        bullet_indent: usize,
        text_indent: usize,
    },

    /// A `Figure` consists of image data (including image options), an optional *caption* (a single paragraph),
    /// and an optional legend (arbitrary body elements).
    /// For page-based output media, figures might float to a different position if this helps the page layout.
    Caption { indent: usize },

    /// A generic citation target.
    Citation { body_indent: usize, label: String },

    /// A reference to a bibliographic citation.
    CitationReference {
        displayed_text: String,
        target_label: String,
    },

    /// A container that triggers the storage of HTML classes
    /// into the DocTree and gives these to its children, if it has any.
    Class {
        body_indent: usize,
        classes: Vec<String>,
    },

    /// A classifier for a `DefinitionTerm` in a `DefinitionList`.
    /// Could be the type of a varible in a function decraration, or something similar.
    Classifier,

    /// The "code" directive constructs a literal block.
    /// If the code language is specified, the content is parsed by the Pygments syntax highlighter
    /// and tokens are stored in nested inline elements with class arguments according to their syntactic category.
    /// The actual highlighting requires a style-sheet (e.g. one generated by Pygments, see the sandbox/stylesheets
    /// for examples).
    ///
    /// The parsing can be turned off with the syntax_highlight configuration setting and command line option or by
    /// specifying the language as :class: option instead of directive argument. This also avoids warnings when Pygments
    /// is not installed or the language is not in the supported languages and markup formats.
    ///
    /// For inline code, use the "code" role.
    ///
    /// Details: https://docutils.sourceforge.io/docs/ref/rst/directives.html#code
    Code {
        text: String,

        language: Option<String>,
        name: Option<String>,
        class: Option<String>,
        number_lines: Option<String>,
    },

    /// TODO
    ColSpec,

    /// A reStructuredText comment
    Comment { text: Option<String> },

    /// The "compound" directive is used to create a compound paragraph,
    /// which is a single logical paragraph containing multiple physical body elements
    /// such as simple paragraphs,literal blocks, tables, lists, etc.,
    /// instead of directly containing text and inline elements. For example:
    ///
    /// Details: https://docutils.sourceforge.io/docs/ref/rst/directives.html#compound-paragraph
    CompoundParagraph {
        body_indent: usize,

        name: Option<String>,
        class: Option<String>,
    },

    /// TODO
    Contact,

    /// The "container" directive surrounds its contents (arbitrary body elements) with a generic block-level "container" element.
    /// Combined with the optional "classes" attribute argument(s), this is an extension mechanism for users & applications.
    /// The "container" directive is the equivalent of HTML's <div> element. It may be used to group a sequence of elements for user-
    /// or application-specific purposes.
    ///
    /// Details: https://docutils.sourceforge.io/docs/ref/rst/directives.html#container
    Container {
        body_indent: usize,

        class_names: Option<Vec<String>>,
        name: Option<String>,
    },

    /// TODO
    Copyright,

    /// The "csv-table" directive is used to create a table from CSV (comma-separated values) data. CSV is
    /// a common data format generated by spreadsheet applications and commercial databases.
    /// The data may be internal (an integral part of the document) or external (a separate file).
    ///
    /// Details: https://docutils.sourceforge.io/docs/ref/rst/directives.html#id4
    CSVTable {
        name: Option<String>,
        class: Option<String>,
        widths: Option<TableColWidths>,
        width: Option<MetricType>,
        header_rows: Option<u32>,
        stub_columns: Option<u32>,
        header: Option<Vec<String>>,
        file: Option<String>,
        url: Option<String>,
        encoding: Option<String>,
        delim: Option<char>,
        quote: Option<char>,
        keepspace: Option<bool>,
        escape: Option<char>,
        align: Option<HorizontalAlignment>,
    },

    /// Document metadata
    Date,

    /// ?
    Decoration,

    /// The definition of a definition list.
    Definition,

    /// A list of definitions. Contains `DefinitionListItems` or `EmptyLine` nodes
    /// as its direct children.
    DefinitionList { term_indent: usize },

    /// A child node type of `DefinitionList`.
    /// Contains a map of `DefinitionTerm`s and the corresponding
    /// `TermDefinitions`, in addition to optional term classifiers.
    DefinitionListItem {
        term: String,
        classifiers: Vec<String>,
        body_indent: usize,
    },

    /// ?
    Description,

    /// ?
    DocInfo,

    /// These are interactive Python sessions contained in Python docstrings.
    /// Based on the Python standard library [doctest](http://www.python.org/doc/current/lib/module-doctest.html) module.
    ///
    /// Doctest blocks begin with ">>>", the python REPL main prompt and end with a blank line.
    /// They are a special case of the literal block and if both are present,
    /// the literal block takes precedence.
    DoctestBlock,

    /// The root node of an reStructuredText document tree.
    Document,

    /// Emphasised or italicized text.
    Emphasis { text: String },

    /// A simple empty line, that contains no actual data.
    /// These can be contained in pretty much any container
    /// node, such as lists or list items, in addition to
    /// existing between body level elements.
    EmptyLine,

    /// Table data cells or entries may contain arbitrary body elements.
    Entry {
        // Whether this node is the last element in its container row.
        is_last: bool,
    },

    /// An enumerated list node. Can only contain `EnumeratedListItem` and `EmptyLine`
    /// nodes as its direct children.
    EnumeratedList {
        delims: EnumDelims,
        kind: EnumKind,
        start_index: usize,
        n_of_items: usize,
        enumerator_indent: usize,
    },

    /// Child node type of `EnumeratedList`. Can contain any `Body`elements
    /// as its children.
    EnumeratedListItem {
        delims: EnumDelims,
        kind: EnumKind,
        index_in_list: usize,
        enumerator_indent: usize,
        text_indent: usize,
    },

    /// A target for an external hyperlink.
    /// Contains a URI pointing  to an external resource
    ExternalHyperlinkTarget {
        marker_indent: usize,
        target: String,
        uri: String,
    },

    /// A field of a field list
    Field,

    /// The parameter that `FieldName` refers to. May contain arbitrary body elements,
    /// just like bulleted and enumerated list items. The first line after the marker specifies
    /// the indentation used as a reference for parsing the rest of the block.
    FieldBody { indentation: usize },

    /// A list of fields, that are used as a part of the
    /// reStructuredText extension syntax, such as directives.
    /// Bibliographies are a special case of these types of lists.
    FieldList { marker_indent: usize },

    /// A field item of a `FieldList`. Consists of a marker with a field name and a
    /// field body consisting of arbitrary body elements.
    /// ```text
    /// +--------------------+----------------------+
    /// | ":" field name ":" | field body           |
    /// +-------+------------+                      |
    ///         | (body elements)+                  |
    ///         +-----------------------------------+
    /// ```
    FieldListItem {
        raw_marker_name: String,
        marker_name_as_inline_nodes: Vec<TreeNodeType>,
        marker_indent: usize,
        body_indent: usize,
    },

    /// A "figure" consists of image data (including image options),
    /// an optional caption (a single paragraph), and an optional legend (arbitrary body elements).
    /// For page-based output media, figures might float to a different position if this helps the page layout.
    Figure {
        /// The indentation of the caption and legend of this `Figure`.
        body_indent: usize,

        // Options
        align: Option<HorizontalAlignment>, // not passed to contained image
        figwidth: Option<Length>,
        figclass: Option<String>,
        name: Option<String>,
    },

    /// The "header" and "footer" directives create document decorations, useful for page navigation, notes, time/datestamp, etc.
    ///
    /// Details: https://docutils.sourceforge.io/docs/ref/rst/directives.html#document-header-footer
    Footer { body_indent: usize },

    /// A foonote citation target. Contains a label and the foornote text itself.
    ///
    /// The "target-notes" directive creates a footnote for each external target in the text,
    /// and corresponding footnote references after each reference. For every explicit target (of the form, .. _target name: URL) in the text,
    /// a footnote will be generated containing the visible URL as content.
    ///
    /// Details: https://docutils.sourceforge.io/docs/ref/rst/directives.html#target-footnotes
    Footnote {
        body_indent: usize,
        kind: FootnoteKind,
        label: String,  // Displayed label
        target: String, // Reference target
    },

    /// A reference to a foot note.
    FootnoteReference {
        displayed_text: String,
        target_label: String,
        kind: FootnoteKind
    },

    /// ?
    Generated,

    /// The "header" and "footer" directives create document decorations, useful for page navigation, notes, time/datestamp, etc.
    ///
    /// Details: https://docutils.sourceforge.io/docs/ref/rst/directives.html#document-header-footer
    Header { body_indent: usize },

    /// An image without a surrounding figure node. Can be inserted inline, as well
    /// (as soon asn substitution definitions and references get implemented...).
    Image {
        uri: String,
        // Options
        alt: Option<String>,
        height: Option<Length>,
        width: Option<Length>,
        scale: Option<f64>,
        align: Option<HTMLAlignment>,
        target: Option<String>,
        name: Option<String>,
        class: Option<String>,
        /// Images can be inserted inline via substitution references
        /// (not yet implemented in ruSTLa).
        inline: bool
    },

    Include {
        uri: String,
        // Options
        start_line: Option<u32>,
        end_line: Option<u32>,
        start_after: Option<u32>,
        end_before: Option<u32>,
        literal: bool,
        code: Option<Option<String>>,
        /// This might be set or not, and if it is, it might contain a start line number or not.
        number_lines: Option<Option<u32>>,
        encoding: Option<String>,
        tab_width: Option<u32>,
        name: Option<String>,
        class: Option<String>
    },

    /// An indirect hyperlink target. Contains a hyperlink reference pointing
    /// to an internal or and external hyperlink.
    IndirectHyperlinkTarget {
        marker_indent: usize,
        target: String,
        indirect_target: String,
    },

    /// ?
    Inline,

    /// An inline reference target.
    InlineTarget { target_label: String },

    /// This node type is reserved for unknown interpreted text roles.
    InterpretedText {
        role: String,
        content: String
    },

    /// ?
    Label,

    /// A `Figure` consists of image data (including image options), an optional caption (a single paragraph),
    /// and an optional *legend* (arbitrary body elements).
    /// For page-based output media, figures might float to a different position if this helps the page layout.
    Legend { body_indent: usize },

    /// A line of a line block.
    Line,

    /// A block of text where each new line begins with an unindented '|',
    /// followed be text with specific left-alignment, used as a reference
    /// for the rest of the block.
    /// Allows writing blocks of text, where the struture of the lines
    /// is meaningful, such as poetry.
    ///
    /// The symbols '|' may be omitted, as they signify the start of a new
    /// line in the rendered output.
    /// ```txt
    /// +------+-----------------------+
    /// | "| " | line                  |
    /// +------| continuation line     |
    ///        +-----------------------+
    /// ```
    LineBlock,

    /// The "list-table" directive is used to create a table from data in a uniform two-level bullet list.
    /// "Uniform" means that each sublist (second-level list) must contain the same number of list items.
    ///
    /// Details: https://docutils.sourceforge.io/docs/ref/rst/directives.html#list-table
    ListTable {
        body_indent: usize,

        /// An optional title, given as the directive argument.
        title: Option<String>,
        /// A comma- or space-separated list of relative column widths. The default is equal-width columns (100%/#columns).
        /// The special value "auto" may be used by writers to decide whether to delegate the determination of column widths to the backend (LaTeX, the HTML browser, ...).
        widths: Option<TableColWidths>,
        /// Forces the width of the table to the specified length or percentage of the line width. If omitted, the renderer determines the width of the table based on its contents.
        width: Option<MetricType>,
        /// The number of rows of list data to use in the table header. Defaults to 0.
        header_rows: Option<u32>,
        /// The number of table columns to use as stubs (row titles, on the left). Defaults to 0.
        stub_columns: Option<u32>,
        /// The horizontal alignment of the table. (New in Docutils 0.13)
        align: Option<HorizontalAlignment>,
    },

    /// Literal text, usually reserved for code.
    Literal { text: String },

    /// Paragraph (possibly empty) ending in a "::" signifies the start of a literal block of text.
    /// Text contained in a literal block is not interpreted in any way,
    /// but simply stored in this node as is.
    LiteralBlock { text: String },

    /// The "math" directive inserts blocks with mathematical content (display formulas, equations)
    /// into the document. The input format is subset of LaTeX math syntax with support for Unicode symbols.
    /// For inline formulas, use the "math" role.
    ///
    /// Details: https://docutils.sourceforge.io/docs/ref/rst/directives.html#math
    Math {
        text: String,
        name: Option<String>,
        class: Option<String>,
    },

    /// A node for display-style mathematics (LaTeX).
    MathBlock {
        math_block: String,

        name: Option<String>,
        class: Option<String>,
    },

    /// A two-column list of command line options, such as the ones typically seen on unix `man` pages.
    /// Four types of options are supported:
    ///
    /// 1. short POSIX options with one '-' and an opion letter,
    /// 2. Long POSIX options with "--", followed by an option word.
    ///    Some systems might use a single dash.
    /// 3. Old GNU-style options starting with a '+', followed by an option letter (!!!deprecated!!!)
    /// 4. DOS/VMS options starting with a '/', followed by an option letter or a word.
    ///
    /// The recognized syntax is based on Python's `getopt.py` module.
    OptionList { option_indent: usize },

    /// A single option in an `OptionList`. Consists of an option,
    /// folllowed by and optional argument and a description.
    /// May contain arbitrary indented body elements after these:
    /// ```text
    /// +----------------------------+-------------+
    /// | option [" " argument] "  " | description |
    /// +-------+--------------------+             |
    ///         | (body elements)+                 |
    ///         +----------------------------------+
    /// ```
    OptionListItem { body_indent: usize },

    /// ?
    OptionString,

    /// Document metadata
    Organization,

    /// A node constructed of a left-aligned block of text
    /// with no special starter markers.
    Paragraph { indent: usize },

    ParsedLiteralBlock,

    Pending,

    Problematic,

    Raw,

    /// A general reference to a reference target.
    Reference {
        displayed_text: Option<String>,
        reference: crate::common::Reference,
    },

    Revision,

    Row,

    /// The "rubric" directive inserts a "rubric" element into the document tree. A rubric is like an informal
    /// heading that doesn't correspond to the document's structure.
    ///
    /// Details: https://docutils.sourceforge.io/docs/ref/rst/directives.html#rubric
    Rubric {
        name: Option<String>,
        class: Option<String>,
    },

    /// A section title node, that contains the title text,
    /// in addition to its marker type and (sub)section level.
    Section {
        title_text: String,
        level: usize,
        line_style: SectionLineStyle,
    },

    /// Sidebars are like miniature, parallel documents that occur inside other documents, providing related or reference material.
    /// A sidebar is typically offset by a border and "floats" to the side of the page; the document's main text may flow around it.
    /// Sidebars can also be likened to super-footnotes; their content is outside of the flow of the document's main text.
    ///
    /// The sidebar element is a nonrecursive section-like construct which may occur at the top level of a section wherever a body element
    /// (list, table, etc.) is allowed. In other words, sidebar elements cannot nest inside body elements, so you can't have a sidebar inside
    /// a table or a list, or inside another sidebar (or topic).
    Sidebar { body_indent: usize },

    Status,

    /// Strongly emphasised text, usually rendered in bold.
    StrongEmphasis { text: String },

    /// As the name impies, generates inline subscripted text.
    Subscript { text: String },

    /// Explicit markup node, as in begins with ".. " followed by a vertical bar '|',
    /// substitution text and another '|'. The text may not begin or end with whitespace.
    /// Substitution definition blocks may contain a nested, *inline compatible* directive
    /// *without* the leading ".. ", such as `image` or `replace`.
    /// ```text
    /// +-------+-----------------------------------------------------+
    /// | ".. " | "|" substitution text "| " directive type "::" data |
    /// +-------+ directive block                                     |
    ///         |                                                     |
    ///         +-----------------------------------------------------+
    /// ```
    SubstitutionDefinition,

    /// A reference that is to be substituted with the reference target directive output.
    SubstitutionReference {
        substitution_label: String,
        // displayed_text: Option<String>,
        target_label: Option<String>,
    },

    /// Document metadata
    Subtitle,

    /// Inline superscripted text.
    Superscript { text: String },

    /// An error node (currently unused).
    SystemMessage,

    /// The "table" directive is used to associate a title with a table or specify options.
    ///
    /// Details: https://docutils.sourceforge.io/docs/ref/rst/directives.html#table
    Table {
        name: Option<String>,
        class: Option<String>,
        align: Option<HorizontalAlignment>,
        widths: Option<Vec<usize>>,
        width: Option<Length>,
    },

    Target,
    /// A table body node
    TBody,

    Term,

    /// A plain text node, that contains no special markup.
    Text { text: String },

    /// Specifies atable group
    TGroup,

    /// Specifies a table header.
    THead,

    /// Specifies a table row.
    TRow,

    /// Document metadata.
    Title,

    /// A reference to a title.
    TitleReference {
        displayed_text: String,
        target_label: String,
    },

    /// The topic element is a nonrecursive section-like construct which may occur at the top level of a section wherever a body element (list, table, etc.) is allowed.
    /// In other words, topic elements cannot nest inside body elements, so you can't have a topic inside a table or a list, or inside another topic.
    /// Topics may contain only body elements.
    ///
    /// The "contents" directive generates a table of contents (TOC) in a topic.
    /// Topics, and therefore tables of contents, may occur anywhere a section or transition may occur.
    /// Body elements and topics may not contain tables of contents.
    ///
    /// Details: https://docutils.sourceforge.io/docs/ref/rst/directives.html#table-of-contents
    Topic {
        title: String,
        depth: Option<u32>,
        local: Option<bool>,
        backlinks: Option<ToCBacklinks>,
        class: Option<String>,
    },

    /// A node corresponding to LaTeX's `\hrulefill` command.
    Transition,

    UnknownDirective {
        directive_name: String,
        argument: String,
        options: std::collections::HashMap<String, String>,
        body_indent: usize,
    },

    /// Document metadata
    Version,

    /// Generic inline whitespace that covers everything from spaces to newlines.
    WhiteSpace { text: String },

    // ============================
    //  Sphinx specific directives
    // ============================
    SphinxOnly {
        expression: String,
        body_indent: usize,
    },

    /// A Sphinx-specific code block. See https://www.sphinx-doc.org/en/master/usage/restructuredtext/directives.html#directive-code-block
    /// for details.
    SphinxCodeBlock {
        language: String,
        linenos: bool,
        lineno_start: Option<usize>,
        emphasize_lines: Option<Vec<usize>>,
        caption: Option<String>,
        name: Option<String>,
        dedent: Option<usize>,
        force: bool,
        code_text: String,
    },

    // ========================
    //  A+ specific directives
    // ========================

    /// A "point of interest" summary block.
    /// A point of interest is mostly like a normal admonition
    /// ("coloured info box"), but they are also linked to each other
    /// with next/previous links. The links enable the user to quickly
    /// navigate between the points of interest.
    AplusPOI {
        title: String,

        // Information needed by the parser
        body_indent: usize,

        // Options
        id: Option<String>,
        previous: Option<String>,
        next: Option<String>,
        hidden: Option<String>,
        class: Option<String>,
        height: Option<Length>,
        columns: Option<String>,
        bgimg: Option<String>,
        not_in_slides: Option<String>,
        not_in_book: Option<String>,
        no_poi_box: Option<String>,
    },

    /// A node that can be inserted inside A+ directives that support multiple columns.
    /// Placing a line that conforms to he regex `[ ]+::newcol` with the `::newcol` having proper indentation
    /// with respect to the multicol directive will generate this type of node.
    AplusColBreak,

    /// A quiz node that corresponds to the questionnaire A+ directive.
    AplusQuestionnaire {
        // Required by the parser
        body_indent: usize,
        points_from_children: QuizPoints,

        // Directive arguments
        key: String,
        max_points: Option<QuizPoints>,
        difficulty: Option<String>,

        // Directive options
        // source: https://github.com/apluslms/a-plus-rst-tools#1-graded-questionnaire
        submissions: Option<String>,
        points_to_pass: Option<String>,
        feedback: Option<String>,
        title: Option<String>,
        no_override: Option<String>,
        pick_randomly: Option<String>,
        preserve_questions_between_attempts: Option<String>,
        category: Option<String>,
        status: Option<String>,
        reveal_model_at_max_submissions: Option<String>,
        show_model: Option<String>,
        allow_assistant_viewing: Option<String>,
        allow_assistant_grading: Option<String>,
    },

    /// See the documentation for the `quoestionnaire` directive:
    /// https://github.com/apluslms/a-plus-rst-tools#1-graded-questionnaire
    AplusPickOne {
        // Needed by the parser
        body_indent: usize,

        points: QuizPoints,
        class: Option<String>,
        required: bool,
        key: Option<String>,
        dropdown: bool,
    },

    /// See the documentation for the `quoestionnaire` directive:
    /// https://github.com/apluslms/a-plus-rst-tools#1-graded-questionnaire
    AplusPickAny {
        // Needed by the parser
        body_indent: usize,

        points: QuizPoints,
        class: Option<String>,
        required: bool,
        key: Option<String>,
        partial_points: bool,
        randomized: bool,
        correct_count: Option<u32>,
        preserve_questions_between_attempts: bool,
    },

    /// See the documentation for the `quoestionnaire` directive:
    /// https://github.com/apluslms/a-plus-rst-tools#1-graded-questionnaire
    AplusFreeText {
        // Needed by the parser
        body_indent: usize,

        // Directive arguments
        points: QuizPoints,

        /// One of `int`, `float`, `string`, `subdiff`, `regexp`, `unsortedchars`,
        /// with `string` and `regexp` having the `-`-separated modifiers
        /// , `ignorews`, `ignorequotes`, `requirecase`, `ignorerepl` and `ignoreparenthesis` available.
        compare_method: String,

        model_answer: String,

        // Directive options
        class: Option<String>,
        required: Option<String>,
        key: Option<String>,
        length: Option<String>,
        height: Option<String>,
    },

    /// A node that contains the choices of an A+ `pick` type question.
    AplusPickChoices { body_indent: usize },

    /// A single answer choice inside an A+ pick-(one|any) question.
    AplusPickChoice {
        label: String,
        is_correct: bool,
        is_pre_selected: bool,
        is_neutral: bool, // Only set with pick-any nodes
    },

    /// A list type node for containing `AplusQuestionnaireHint`s.
    AplusQuestionnaireHints { body_indent: usize },

    /// A hint for questionnaire pick-{one, any} and freetext questions.
    AplusQuestionnaireHint {
        label: String,
        show_when_not_selected: bool,
        question_type: crate::common::AplusQuestionnaireType,
    },

    /// A+ submit-type exercise. See https://github.com/apluslms/a-plus-rst-tools#3-submittable-exercise.
    AplusSubmit {
        body_indent: usize,
        key: String,
        difficulty: String,
        max_points: QuizPoints,

        config: String,
        submissions: u32,
        points_to_pass: QuizPoints,
        class: String,
        title: String,
        category: String, // Default: submit
        status: AplusExerciseStatus,
        ajax: bool,
        allow_assistant_viewing: bool,
        allow_assistant_grading: bool,
        quiz: bool,
        url: String,
        radar_tokenizer: AplusRadarTokenizer,
        radar_minimum_match_tokens: Option<u32>,
        lti: String,
        lti_resource_link_id: String,
        lti_open_in_iframe: bool,
        lti_aplus_get_and_post: bool,
    },

    AplusActiveElementInput {
        key_for_input: String,
        title: Option<String>,
        default: Option<String>,
        class: Option<String>,
        width: Option<Length>,
        height: Option<Length>,
        clear: Option<AplusActiveElementClear>,
        input_type: Option<AplusActiveElementInputType>,
        file: Option<String>,
    },

    AplusActiveElementOutput {
        key_for_output: String,
        config: String,
        inputs: String,
        title: Option<String>,
        class: Option<String>,
        width: Option<Length>,
        height: Option<Length>,
        clear: Option<AplusActiveElementClear>,
        output_type: AplusActiveElementOutputType,
        submissions: Option<u32>,
        scale_size: bool,
        status: AplusExerciseStatus,
    },
}

use crate::doctree::node_categories::*;

impl TreeNodeType {

    /// Returns the relevant content indent of each node type.
    /// This is useful, when one is trying to find out whether a node belongs to a parent node.
    /// Node containerzation is determined by indentation in reStructuredText.
    pub fn body_indent(&self) -> Option<usize> {
        match self {
            Self::Abbreviation { .. } => None,
            Self::AbsoluteURI { .. } => None,
            Self::Acronym { .. } => None,
            Self::Address => None,
            Self::Admonition { content_indent, .. } => Some(*content_indent),
            Self::Attribution { .. } => None,
            Self::Author { .. } => None,
            Self::Authors { .. } => None,
            Self::AutomaticSectionNumbering { .. } => None,
            Self::BlockQuote { body_indent, .. } => Some(*body_indent),
            Self::BulletList { bullet_indent, .. } => Some(*bullet_indent),
            Self::BulletListItem { text_indent, .. } => Some(*text_indent),
            Self::Caption { .. } => None,
            Self::Citation { body_indent, .. } => Some(*body_indent),
            Self::CitationReference { .. } => None,
            Self::Class { body_indent, .. } => Some(*body_indent),
            Self::Classifier { .. } => None,
            Self::Code { .. } => None,
            Self::ColSpec { .. } => None,
            Self::Comment { .. } => None,
            Self::CompoundParagraph { body_indent, .. } => Some(*body_indent),
            Self::Contact { .. } => None,
            Self::Container { body_indent, .. } => Some(*body_indent),
            Self::Copyright { .. } => None,
            Self::CSVTable { .. } => None,
            Self::Date => None,
            Self::Decoration => None,
            Self::Definition => None,
            Self::DefinitionList { term_indent, .. } => Some(*term_indent),
            Self::DefinitionListItem { body_indent, .. } => Some(*body_indent),
            Self::Description => None,
            Self::DocInfo => None,
            Self::DoctestBlock { .. } => None,
            Self::Document { .. } => Some(0),
            Self::Emphasis { .. } => None,
            Self::EmptyLine => None,
            Self::Entry { .. } => None,
            Self::EnumeratedList {
                enumerator_indent, ..
            } => Some(*enumerator_indent),
            Self::EnumeratedListItem { text_indent, .. } => Some(*text_indent),
            Self::ExternalHyperlinkTarget { .. } => None,
            Self::Field => None,
            Self::FieldBody { .. } => None,
            Self::FieldList { marker_indent, .. } => Some(*marker_indent),
            Self::FieldListItem { body_indent, .. } => Some(*body_indent),
            Self::Figure { body_indent, .. } => Some(*body_indent),
            Self::Footer { body_indent, .. } => Some(*body_indent),
            Self::Footnote { body_indent, .. } => Some(*body_indent),
            Self::FootnoteReference { .. } => None,
            Self::Header { body_indent, .. } => Some(*body_indent),
            Self::Generated => None,
            Self::Image { .. } => None,
            Self::Include { .. } => None,
            Self::IndirectHyperlinkTarget { .. } => None,
            Self::Inline { .. } => None,
            Self::InlineTarget { .. } => None,
            Self::InterpretedText { .. } => None,
            Self::Label { .. } => None,
            Self::Legend { body_indent, .. } => Some(*body_indent),
            Self::Line { .. } => None,
            Self::LineBlock { .. } => None,
            Self::ListTable { body_indent, .. } => Some(*body_indent),
            Self::Literal { .. } => None,
            Self::LiteralBlock { .. } => None,
            Self::Math { .. } => None,
            Self::MathBlock { .. } => None,
            Self::OptionList { option_indent, .. } => Some(*option_indent),
            Self::OptionListItem { body_indent, .. } => Some(*body_indent),
            Self::OptionString { .. } => None,
            Self::Organization { .. } => None,
            Self::Paragraph { indent, .. } => Some(*indent),
            Self::ParsedLiteralBlock { .. } => None,
            Self::Pending { .. } => None,
            Self::Problematic { .. } => None,
            Self::Raw { .. } => None,
            Self::Reference { .. } => None,
            Self::Revision { .. } => None,
            Self::Row { .. } => None,
            Self::Rubric { .. } => None,
            Self::Section { .. } => Some(0),
            Self::Sidebar { body_indent, .. } => Some(*body_indent),
            Self::Status { .. } => None,
            Self::StrongEmphasis { .. } => None,
            Self::Subscript { .. } => None,
            Self::SubstitutionDefinition { .. } => None,
            Self::SubstitutionReference { .. } => None,
            Self::Subtitle { .. } => None,
            Self::Superscript { .. } => None,
            Self::SystemMessage { .. } => None,
            Self::Table { .. } => None,
            Self::Target { .. } => None,
            Self::TBody { .. } => None,
            Self::Term { .. } => None,
            Self::Text { .. } => None,
            Self::TGroup { .. } => None,
            Self::THead { .. } => None,
            Self::TRow { .. } => None,
            Self::Title { .. } => None,
            Self::TitleReference { .. } => None,
            Self::Topic { .. } => None,
            Self::Transition => None,
            Self::UnknownDirective { body_indent, .. } => Some(*body_indent),
            Self::Version { .. } => None,
            Self::WhiteSpace { .. } => None,

            // ========================
            //  Sphinx specific directives
            // ========================
            Self::SphinxOnly { body_indent, .. } => Some(*body_indent),
            Self::SphinxCodeBlock { .. } => None,

            // ========================
            //  A+ specific directives
            // ========================
            Self::AplusPOI { body_indent, .. } => Some(*body_indent),
            Self::AplusColBreak => None,
            Self::AplusQuestionnaire { body_indent, .. } => Some(*body_indent),
            Self::AplusPickOne { body_indent, .. } => Some(*body_indent),
            Self::AplusPickAny { body_indent, .. } => Some(*body_indent),
            Self::AplusFreeText { body_indent, .. } => Some(*body_indent),
            Self::AplusPickChoices { body_indent, .. } => Some(*body_indent),
            Self::AplusPickChoice { .. } => None,
            Self::AplusQuestionnaireHints { body_indent } => Some(*body_indent),
            Self::AplusQuestionnaireHint { .. } => None,
            Self::AplusSubmit { body_indent, .. } => Some(*body_indent),
            Self::AplusActiveElementInput { .. } => None,
            Self::AplusActiveElementOutput { .. } => None,
        }
    }

    /// According to the [reStructuredText DTD](https://docutils.sourceforge.io/docs/ref/doctree.html)
    /// each node type is associated with a set of categories.
    /// This function returns that set (an immutable iterator) for each defined node type.
    pub fn node_categories(&self) -> impl Iterator<Item = &NodeCategory> {
        let categories: &[NodeCategory] = match self {
            Self::Abbreviation { .. } => &ABBREVIATION_CATEGORIES,
            Self::AbsoluteURI { .. } => &ABSOLUTE_URI_CATEGORIES,
            Self::Acronym { .. } => &ACRONYM_CATEGORIES,
            Self::Address => &ADDRESS_CATEGORIES,
            Self::Admonition { .. } => &ADMONITION_CATEGORIES,
            Self::Attribution { .. } => &ATTRIBUTION_CATEGORIES,
            Self::Author { .. } => &AUTHOR_CATEGORIES,
            Self::Authors { .. } => &AUTHORS_CATEGORIES,
            Self::AutomaticSectionNumbering { .. } => &AUTO_SECTION_NUMBERING_CATEGORIES,
            Self::BlockQuote { .. } => &BLOCK_QUOTE_CATEGORIES,
            Self::BulletList { .. } => &BULLET_LIST_CATEGORIES,
            Self::BulletListItem { .. } => &BULLET_LIST_ITEM_CATEGORIES,
            Self::Caption { .. } => &CAPTION_CATEGORIES,
            Self::Citation { .. } => &CITATION_CATEGORIES,
            Self::CitationReference { .. } => &CITATION_REFERENCE_CATEGORIES,
            Self::Class { .. } => &CLASS_CATEGORIES,
            Self::Classifier { .. } => &CLASSIFIER_CATEGORIES,
            Self::Code { .. } => &CODE_CATEGORIES,
            Self::ColSpec { .. } => &COLSPEC_CATEGORIES,
            Self::Comment { .. } => &COMMENT_CATEGORIES,
            Self::CompoundParagraph { .. } => &COMPOUND_PARAGRAPH_CATEGORIES,
            Self::Contact { .. } => &CONTACT_CATEGORIES,
            Self::Container { .. } => &CONTAINER_CATEGORIES,
            Self::Copyright { .. } => &COPYRIGHT_CATEGORIES,
            Self::CSVTable { .. } => &CSV_TABLE_CATEGORIES,
            Self::Date => &DATE_CATEGORIES,
            Self::Decoration => &DECORATION_CATEGORIES,
            Self::Definition => &DEFINITION_CATEGORIES,
            Self::DefinitionList { .. } => &DEFINITION_LIST_CATEGORIES,
            Self::DefinitionListItem { .. } => &DEFINITION_LIST_ITEM_CATEGORIES,
            Self::Description => &DESCRIPTION_CATEGORIES,
            Self::DocInfo => &DOC_INFO_CATEGORIES,
            Self::DoctestBlock { .. } => &DOCTEST_BLOCK_CATEGORIES,
            Self::Document { .. } => &DOCUMENT_CATEGORIES,
            Self::Emphasis { .. } => &EMPHASIS_CATEGORIES,
            Self::EmptyLine => &EMPTY_LINE_CATEGORIES,
            Self::Entry { .. } => &ENTRY_CATEGORIES,
            Self::EnumeratedList { .. } => &ENUMERATED_LIST_CATEGORIES,
            Self::EnumeratedListItem { .. } => &ENUMERATED_LIST_ITEM_CATEGORIES,
            Self::ExternalHyperlinkTarget { .. } => &EXTERNAL_HYPERLINK_TARGET_CATEGORIES,
            Self::Field => &FIELD_CATEGORIES,
            Self::FieldBody { .. } => &FIELD_BODY_CATEGORIES,
            Self::FieldList { .. } => &FIELD_LIST_CATEGORIES,
            Self::FieldListItem { .. } => &FIELD_LIST_ITEM_CATEGORIES,
            Self::Figure { .. } => &FIGURE_CATEGORIES,
            Self::Footer { .. } => &FOOTER_CATEGORIES,
            Self::Footnote { .. } => &FOOTNOTE_CATEGORIES,
            Self::FootnoteReference { .. } => &FOOTNOTE_REFERENCE_CATEGORIES,
            Self::Header { .. } => &HEADER_CATEGORIES,
            Self::Generated => &GENERATED_CATEGORIES,
            Self::Image { .. } => &IMAGE_CATEGORIES,
            Self::Include { .. } => &INCLUDE_CATEGORIES,
            Self::IndirectHyperlinkTarget { .. } => &INDIRECT_HYPERLINK_TARGET_CATEGORIES,
            Self::Inline { .. } => &INLINE_CATEGORIES,
            Self::InlineTarget { .. } => &INLINE_TARGET_CATEGORIES,
            Self::InterpretedText { .. } => &INTERPRETED_TEXT_CATEGORIES,
            Self::Label { .. } => &LABEL_CATEGORIES,
            Self::Legend { .. } => &LEGEND_CATEGORIES,
            Self::Line { .. } => &LINE_CATEGORIES,
            Self::LineBlock { .. } => &LINE_BLOCK_CATEGORIES,
            Self::ListTable { .. } => &LIST_TABLE_CATEGORIES,
            Self::Literal { .. } => &LITERAL_CATEGORIES,
            Self::LiteralBlock { .. } => &LITERAL_BLOCK_CATEGORIES,
            Self::Math { .. } => &MATH_CATEGORIES,
            Self::MathBlock { .. } => &MATH_BLOCK_CATEGORIES,
            Self::OptionList { .. } => &OPTION_LIST_CATEGORIES,
            Self::OptionListItem { .. } => &OPTION_LIST_ITEM_CATEGORIES,
            Self::OptionString { .. } => &OPTION_STRING_CATEGORIES,
            Self::Organization { .. } => &ORGANIZATION_CATEGORIES,
            Self::Paragraph { .. } => &PARAGRAPH_CATEGORIES,
            Self::ParsedLiteralBlock { .. } => &PARSED_LITERAL_BLOCK_CATEGORIES,
            Self::Pending { .. } => &PENDING_CATEGORIES,
            Self::Problematic { .. } => &PROBLEMATIC_CATEGORIES,
            Self::Raw { .. } => &RAW_CATEGORIES,
            Self::Reference { .. } => &REFERENCE_CATEGORIES,
            Self::Revision { .. } => &REVISION_CATEGORIES,
            Self::Row { .. } => &ROW_CATEGORIES,
            Self::Rubric { .. } => &RUBRIC_CATEGORIES,
            Self::Section { .. } => &SECTION_CATEGORIES,
            Self::Sidebar { .. } => &SIDEBAR_CATEGORIES,
            Self::Status { .. } => &STATUS_CATEGORIES,
            Self::StrongEmphasis { .. } => &STRONG_EMPHASIS_CATEGORIES,
            Self::Subscript { .. } => &SUBSCRIPT_CATEGORIES,
            Self::SubstitutionDefinition { .. } => &SUBSTITUTION_DEF_CATEGORIES,
            Self::SubstitutionReference { .. } => &SUBSTITUTION_REF_CATEGORIES,
            Self::Subtitle { .. } => &SUBTITLE_CATEGORIES,
            Self::Superscript { .. } => &SUPERSCRIPT_CATEGORIES,
            Self::SystemMessage { .. } => &SYSTEM_MESSAGE_CATEGORIES,
            Self::Table { .. } => &TABLE_CATEGORIES,
            Self::Target { .. } => &TARGET_CATEGORIES,
            Self::TBody { .. } => &T_BODY_CATEGORIES,
            Self::Term { .. } => &TERM_CATEGORIES,
            Self::Text { .. } => &TEXT_CATEGORIES,
            Self::TGroup { .. } => &T_GROUP_CATEGORIES,
            Self::THead { .. } => &T_HEAD_CATEGORIES,
            Self::TRow { .. } => &T_ROW_CATEGORIES,
            Self::Title { .. } => &TITLE_CATEGORIES,
            Self::TitleReference { .. } => &TITLE_REF_CATEGORIES,
            Self::Topic { .. } => &TOPIC_CATEGORIES,
            Self::Transition {} => &TRANSITION_CATEGORIES,
            Self::UnknownDirective { .. } => &UNKNOWN_DIRECTIVE_CATEGORIES,
            Self::Version { .. } => &VERSION_CATEGORIES,
            Self::WhiteSpace { .. } => &WHITESPACE_CATEGORIES,

            // ============================
            //  Sphinx specific directives
            // ============================
            Self::SphinxOnly { body_indent, .. } => &SPHINX_ONLY_CATEGORIES,
            Self::SphinxCodeBlock { .. } => &SPHINX_CODE_BLOCK_CATEGORIES,

            // ========================
            //  A+ specific directives
            // ========================
            Self::AplusPOI { .. } => &APLUS_POI_CATEGORIES,
            Self::AplusColBreak => &APLUS_COL_BREAK_CATEGORIES,
            Self::AplusQuestionnaire { .. } => &APLUS_QUESTIONNAIRE_CATEGORIES,
            Self::AplusPickOne { .. } => &APLUS_PICK_ONE_CATEGORIES,
            Self::AplusPickAny { .. } => &APLUS_PICK_ANY_CATEGORIES,
            Self::AplusFreeText { .. } => &APLUS_FREETEXT_CATEGORIES,
            Self::AplusPickChoices { body_indent, .. } => &APLUS_PICK_CHOICES_CATEGORIES,
            Self::AplusPickChoice { .. } => &APLUS_PICK_CHOICE_CATEGORIES,
            Self::AplusQuestionnaireHints { .. } => &APLUS_QUESTIONNAIRE_HINTS_CATEGORIES,
            Self::AplusQuestionnaireHint { .. } => &APLUS_QUESTIONNAIRE_HINT_CATEGORIES,
            Self::AplusSubmit { .. } => &APLUS_SUBMIT_CATEGORIES,
            Self::AplusActiveElementInput { .. } => &APLUS_ACTIVE_ELEMENT_INPUT_CATEGORIES,
            Self::AplusActiveElementOutput { .. } => &APLUS_ACTIVE_ELEMENT_OUTPUT_CATEGORIES,
        };

        categories.iter()
    }
}

impl std::fmt::Display for TreeNodeType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let node_type_str: &str = match self {
            Self::Abbreviation { .. } => "abbreviation",
            Self::AbsoluteURI { .. } => "absolute URI",
            Self::Acronym { .. } => "acronym",
            Self::Address => "address",
            Self::Admonition { .. } => "admonition",
            Self::Attribution { .. } => "attribution",
            Self::Author { .. } => "author",
            Self::Authors { .. } => "authors",
            Self::AutomaticSectionNumbering { .. } => {
                todo!("auto section numbering is not really a node?")
            }
            Self::BlockQuote { .. } => "block quote",
            Self::BulletList { .. } => "bullet list",
            Self::BulletListItem { .. } => "bullet list item",
            Self::Caption { .. } => "caption",
            Self::Citation { .. } => "citation",
            Self::CitationReference { .. } => "citation reference",
            Self::Class { .. } => "class",
            Self::Classifier { .. } => "classifier",
            Self::Code { .. } => "colspec",
            Self::ColSpec { .. } => "colspec",
            Self::Comment { .. } => "comment",
            Self::CompoundParagraph { .. } => "compound paragraph",
            Self::Contact { .. } => "contact",
            Self::Container { .. } => "container",
            Self::Copyright { .. } => "copyright",
            Self::CSVTable { .. } => "csv table",
            Self::Date => "date",
            Self::Decoration => "decoration",
            Self::Definition => "definition",
            Self::DefinitionList { .. } => "definition list",
            Self::DefinitionListItem { .. } => "definition list item",
            Self::Description => "description",
            Self::DocInfo => "docinfo",
            Self::DoctestBlock { .. } => "doc test block",
            Self::Document { .. } => "document",
            Self::Emphasis { .. } => "emphasis",
            Self::EmptyLine => "emptyline",
            Self::Entry { .. } => "entry",
            Self::EnumeratedList { .. } => "enumerated list",
            Self::EnumeratedListItem { .. } => "enumerated list item",
            Self::ExternalHyperlinkTarget { .. } => "external hyperlink target",
            Self::Field => "field",
            Self::FieldBody { .. } => "field body",
            Self::FieldList { .. } => "field list",
            Self::FieldListItem { .. } => "field list item",
            Self::Figure { .. } => "figure",
            Self::Footer { .. } => "footer",
            Self::Footnote { .. } => "footnote",
            Self::FootnoteReference { .. } => "footnote reference",
            Self::Header { .. } => "header",
            Self::Generated => "generated",
            Self::Image { .. } => "image",
            Self::Include { .. } => "include",
            Self::IndirectHyperlinkTarget { .. } => "hyperlink target",
            Self::Inline { .. } => "inline",
            Self::InlineTarget { .. } => "inline target",
            Self::InterpretedText { .. } => "interpreted text",
            Self::Label { .. } => "label",
            Self::Legend { .. } => "legend",
            Self::Line { .. } => "line",
            Self::LineBlock { .. } => "line block",
            Self::ListTable { .. } => "list table",
            Self::Literal { .. } => "literal",
            Self::LiteralBlock { .. } => "literal block",
            Self::Math { .. } => "math",
            Self::MathBlock { .. } => "math block",
            Self::OptionList { .. } => "option list",
            Self::OptionListItem { .. } => "option list item",
            Self::OptionString { .. } => "option string",
            Self::Organization { .. } => "organization",
            Self::Paragraph { .. } => "paragraph",
            Self::ParsedLiteralBlock { .. } => "parsed literal block",
            Self::Pending { .. } => "pending",
            Self::Problematic { .. } => "problematic",
            Self::Raw { .. } => "raw",
            Self::Reference { .. } => "reference",
            Self::Revision { .. } => "revision",
            Self::Row { .. } => "row",
            Self::Rubric { .. } => "rubric",
            Self::Section { .. } => "section",
            Self::Sidebar { .. } => "sidebar",
            Self::Status { .. } => "status",
            Self::StrongEmphasis { .. } => "strong emphasis",
            Self::Subscript { .. } => "subscript",
            Self::SubstitutionDefinition { .. } => "substitution definition",
            Self::SubstitutionReference { .. } => "substitution reference",
            Self::Subtitle { .. } => "subtitle",
            Self::Superscript { .. } => "superscript",
            Self::SystemMessage { .. } => "system message",
            Self::Table { .. } => "table",
            Self::Target { .. } => "target",
            Self::TBody { .. } => "table body",
            Self::Term { .. } => "term",
            Self::Text { .. } => "text",
            Self::TGroup { .. } => "table group",
            Self::THead { .. } => "table head",
            Self::TRow { .. } => "table row",
            Self::Title { .. } => "title",
            Self::TitleReference { .. } => "title reference",
            Self::Topic { .. } => "topic",
            Self::Transition {} => "transition",
            Self::UnknownDirective { .. } => "unknown directive",
            Self::Version { .. } => "version",
            Self::WhiteSpace { .. } => "whitespace",

            // ============================
            //  Sphinx specific directives
            // ============================
            Self::SphinxOnly { body_indent, .. } => "sphinx only",
            Self::SphinxCodeBlock { .. } => "sphinx code block",

            // ========================
            //  A+ specific directives
            // ========================
            Self::AplusPOI { .. } => "aplus point of interest",
            Self::AplusColBreak => "aplus column break",
            Self::AplusQuestionnaire { .. } => "aplus questionnaire",
            Self::AplusPickOne { .. } => "aplus pick one",
            Self::AplusPickAny { .. } => "aplus pick any",
            Self::AplusFreeText { .. } => "aplus freetext",
            Self::AplusPickChoices { body_indent, .. } => "aplus pick choices",
            Self::AplusPickChoice { .. } => "aplus pick choice",
            Self::AplusQuestionnaireHints { .. } => "aplus questionnaire hints",
            Self::AplusQuestionnaireHint { .. } => "aplus questionnaire hint",
            Self::AplusSubmit { .. } => "aplus submit",
            Self::AplusActiveElementInput { .. } => "aplus active element input",
            Self::AplusActiveElementOutput { .. } => "aplus active element output",
        };
        write!(f, "\"{}\"", node_type_str)
    }
}
